from __future__ import absolute_import

import argparse
import os

import sys
import yaml
from os.path import exists

from phigaro.helper import setup, MetaGeneMarkNotFound, MetaGeneMarkKeyNotFound, HMMERNotFound, download_pvogs, \
    HelperException


def main():
    home = os.getenv('HOME')
    phigaro_dir = os.path.join(home, '.phigaro')
    phigaro_config = os.path.join(phigaro_dir, 'config.yml')
    pvogs_dir = os.path.join(phigaro_dir, 'pvog')

    parser = argparse.ArgumentParser(description="Phigaro setup helper",
                                     formatter_class=argparse.ArgumentDefaultsHelpFormatter)

    parser.add_argument('-c', '--config', default=phigaro_config, help='config path')
    parser.add_argument('-p', '--pvog', default=pvogs_dir, help='pvogs dir')
    # parser.add_argument('-v', '--verbose', action='store_true')

    args = parser.parse_args()

    if exists(args.config):
        print('Phigaro already configured')
        exit(0)

    try:
        config = setup()
    except HelperException as ex:
        sys.stdout.write(ex.message+'\n')
        exit(1)

    config['hmmer']['pvog_path'] = os.path.join(pvogs_dir, 'allpvoghmms')

    # print("Found MetaGeneMark in: {}".format(config['genemark']['bin']))
    # print("Found MetaGeneMark model in: {}".format(config['genemark']['mod_path']))
    # print("Found MetaGeneMark key in: {}".format(config['genemark']['valid_key']))
    # print("Found HMMER in: {}".format(config['hmmer']['bin']))
    print("HMMER model in: {}".format(config['hmmer']['pvog_path']))

    if not exists(pvogs_dir):
        print('Downloading models')
        download_pvogs('http://download.ripcm.com/phigaro/', pvogs_dir)

    with open(args.config, 'w') as f:
        yaml.safe_dump(config, f, default_flow_style=False)




if __name__ == '__main__':
    main()



from .task.base import AbstractTask

import logging
logger = logging.getLogger(__name__)


def make_tasks_chain(task_classes, sample, first_input, context):
    """
    :type task_classes: list[(str, str, Context)->AbstractTask]
    :type first_input: str
    :type sample: str
    :type context: Context
    :rtype: list[AbstractTask]
    """

    input = first_input
    tasks = []
    for task_class in task_classes:
        task = task_class(input, sample, context)
        input = task.output()
        tasks.append(task)

    return tasks


def run_tasks_chain(tasks_chain):
    """
    :type tasks_chain: list[AbstractTask]
    :rtype: str
    """

    for task in tasks_chain:
        logger.info("Executing {task}. output: {output}".format(
            task=task.task_name,
            output=task.output()
        ))
        task.run()

    return task.output()



import csv
import numpy as np
from itertools import groupby


def read_npn(filename, sep=None, as_dict=True):
    sep = sep or ' '
    with open(filename) as f:
        reader = csv.reader(f, delimiter=sep)
        if as_dict:
            return dict(reader)
        else:
            return reader


def read_coords(filename, sep=None):
    sep = sep or '\t'
    with open(filename) as f:
        reader = csv.reader(f, delimiter=sep)
        res = {}
        for phage, group in groupby(reader, key=lambda x: x[0]):
            res[phage] = sorted((
                (int(begin), int(end))
                for _, begin, end in group
            ), key=lambda x: x[0])
    return res


def convert_npn(phage, ph_sym):
    return np.array([
        int(c == ph_sym)
        for c in phage
    ])



import os
import time
from os.path import join, dirname, exists
from glob import glob
from shutil import copy
import sh
from future.backports.urllib.parse import urljoin

HOME = os.getenv("HOME")

# TODO: refactor this to class
is_updated_db = False


class HelperException(Exception):
    message = None


class MetaGeneMarkNotFound(HelperException):
    message = "Didn't find MetaGeneMark anywhere, " + \
              "please download it from http://topaz.gatech.edu/Genemark/license_download.cgi "


class MetaGeneMarkKeyNotFound(HelperException):
    message = "Please update your MetaGeneMark key file: "+ \
              "http://topaz.gatech.edu/GeneMark/license_download.cgi . " + \
              "You can download the \"key\" file only. "


class HMMERNotFound(HelperException):
    message = "No HMMER found. Please download HMMER package from here: http://hmmer.org/download.html "


def locate(*args, **kwargs):
    # TODO: refactor this to class
    global is_updated_db
    if not is_updated_db:
        try:
            with sh.contrib.sudo:
                sh.updatedb()
                is_updated_db = True
        except sh.ErrorReturnCode_1:
            print('Invalid password')
            exit(1)
    return sh.locate(*args, **kwargs)


def find_gmhmmp_bin():
    mgm_location = sh.which("gmhmmp")
    if mgm_location:
        return mgm_location
    # print("MetaGeneMark is not in the PATH")
    try:
        res = locate("mgm/gmhmmp")
        return next(res).rstrip()
    except sh.ErrorReturnCode_1 as ex:
        raise MetaGeneMarkNotFound()


def find_mgm_mod_file(mgm_dir):
    # TODO: handle no or multiple .mod files in mgm_dir
    return glob(join(mgm_dir, '*.mod'))[0]


def is_gm_key_valid(gm_key_path):
    res = os.stat(gm_key_path)
    days = (time.time() - res.st_mtime) / (60 * 60 * 24)
    return days < 400


def find_gm_key():
    gm_keys = [
        gm_key_path.rstrip()
        for gm_key_path in locate('gm_key')
    ]

    valid_keys = [
        gm_key_path
        for gm_key_path in gm_keys
        if exists(gm_key_path) and is_gm_key_valid(gm_key_path)
    ]
    if not valid_keys:
        raise MetaGeneMarkKeyNotFound()
    return valid_keys[0]


def setup_mgm():
    mgm_location = find_gmhmmp_bin()
    print('Found MetaGeneMark binary in: {}'.format(mgm_location))
    mod_file = find_mgm_mod_file(dirname(mgm_location))
    print('Found MetaGeneMark model in: {}'.format(mod_file))
    gm_key_home = join(HOME, '.gm_key')

    if not exists(gm_key_home) or not is_gm_key_valid(gm_key_home):
        valid_gm_key = find_gm_key()
        if valid_gm_key.endswith(".gz"):
            print('Extracting {} to {}'.format(valid_gm_key, gm_key_home))
            sh.gunzip(valid_gm_key, '-c', _out=gm_key_home)
        else:
            print('Copying {} to {}'.format(valid_gm_key, gm_key_home))
            copy(valid_gm_key, gm_key_home)

    return {
            'bin': mgm_location,
            'mod_path': mod_file,
            # 'valid_key': valid_gm_key,
        }


def setup_hmmer():
    # TODO: pVOG files
    hmmsearch_location = sh.which("hmmsearch")

    if hmmsearch_location is None:
        try:
            # in case hmmscan is in several places
            hs_not_in_path = locate("-r", "/hmmsearch$")
            hmmsearch_location = hs_not_in_path.split('\n')[0]
        except sh.ErrorReturnCode_1:
            raise HMMERNotFound()

    print('Found HMMER binary in: {}'.format(hmmsearch_location))
    return {
            'bin': hmmsearch_location,
            'pvog_path': 'software/hmmer/data/allprofiles.hmm',
            'e_value_threshold': 1.0e-5,
        }


def download_pvogs(base_url, out_dir):
    def download_file(filename):
        url = urljoin(base_url, filename)
        out = join(out_dir, filename)
        print('Downloading {url} to {out}'.format(
            url=url,
            out=out,
        ))
        sh.curl('-o', out, url, _tty_out=True)

    if not exists(out_dir):
        os.makedirs(out_dir)
    download_file('allpvoghmms')
    download_file('allpvoghmms.h3f')
    download_file('allpvoghmms.h3i')
    download_file('allpvoghmms.h3m')
    download_file('allpvoghmms.h3p')


def setup():
    mgm_params = setup_mgm()
    hmmer_params = setup_hmmer()

    return {
        'genemark': {
            'bin': mgm_params['bin'],
            'mod_path': mgm_params['mod_path'],
        },
        'hmmer': {
            'bin': hmmer_params['bin'],
            'pvog_path': hmmer_params['pvog_path'],
            'e_value_threshold': 1.0e-5,
        },
        'phigaro': {
            'window_len': 34,
            'threshold_min': 5.341108,
            'threshold_max': 7.571429,
        }
    }



from __future__ import absolute_import

import argparse
import logging
import multiprocessing
import os
import sys
import uuid
from os.path import join, exists

import yaml

from phigaro.context import Context
from phigaro.scheduling.path import sample_name
from phigaro.scheduling.runner import run_tasks_chain
from phigaro.scheduling.task.gene_mark import GeneMarkTask
from phigaro.scheduling.task.hmmer import HmmerTask
from phigaro.scheduling.task.run_phigaro import RunPhigaroTask


def main():
    default_config_path = join(os.getenv('HOME'), '.phigaro', 'config.yml')
    parser = argparse.ArgumentParser(
        description='Phigaro is a scalable command-line tool for predictions phages and prophages '
                    'from nucleid acid sequences (including metagenomes) and '
                    'is based on phage genes HMMs and a smoothing window algorithm.',
    )
    parser.add_argument('-f', '--fasta-file', help='Assembly scaffolds\contigs or full genomes', required=True)
    parser.add_argument('-c', '--config', default=default_config_path, help='config file')
    parser.add_argument('-v', '--verbose', action='store_true', help='print debug information (for developers)')
    parser.add_argument('-t', '--threads',
                        type=int,
                        default=multiprocessing.cpu_count(),
                        help='num of threads ('
                             'default is num of CPUs={})'.format(multiprocessing.cpu_count()))

    args = parser.parse_args()

    logging.basicConfig(level=logging.INFO if args.verbose else logging.WARN)
    logging.getLogger('sh.command').setLevel(logging.WARN)

    logger = logging.getLogger(__name__)

    if not exists(args.config):
        # TODO: pretty message
        print('Please create config file using phigaro-setup script')
        exit(1)

    with open(args.config) as f:
        logger.info('Using config file: {}'.format(args.config))
        config = yaml.load(f)

    filename = args.fasta_file
    sample = '{}-{}'.format(
        sample_name(filename),
        uuid.uuid4().hex
    )

    Context.initialize(
        sample=sample,
        config=config,
        threads=args.threads,
    )

    gene_mark_task = GeneMarkTask(filename)
    hmmer_task = HmmerTask(gene_mark_task=gene_mark_task)
    run_phigaro_task = RunPhigaroTask(gene_mark_task=gene_mark_task, hmmer_task=hmmer_task)

    output_file = run_tasks_chain([
        gene_mark_task,
        hmmer_task,
        run_phigaro_task
    ])

    with open(output_file) as f:
        for line in f:
            sys.stdout.write(line)


if __name__ == '__main__':
    main()



from builtins import super
import re
import sh
from itertools import groupby
import csv

from .base import AbstractTask
from .gene_mark import GeneMarkTask


class HmmerTask(AbstractTask):
    task_name = 'hmmer'

    def __init__(self, gene_mark_task):
        """

        :type gene_mark_task: GeneMarkTask
        """
        super().__init__()
        self.gene_mark = gene_mark_task

    def _prepare(self):
        self.hmmer = (
            sh.Command(self.config['hmmer']['bin'])
        )
        self._tmp_hmmer_out = self.file('{}.hmmer_out'.format(self.sample))

    def output(self):
        return self.file('{}.npn'.format(self.sample))

    def run(self):
        self.hmmer('--cpu', self.context.threads,
                   '--notextw',
                   '--tblout', self._tmp_hmmer_out,
                   self.config['hmmer']['pvog_path'],
                   self.gene_mark.output())
        self._parse_hmmer_output()

    @staticmethod
    def parse_line(line):
        tokens = re.split(r'\s+', line)
        scaffold = '>' + line.split('>')[1]
        name = tokens[0]
        evalue = float(tokens[4])
        return scaffold, name, evalue

    def _parse_hmmer_output(self):
        max_evalue = self.config['hmmer']['e_value_threshold']

        with open(self._tmp_hmmer_out) as f:
            it = (
                self.parse_line(line.strip())
                for line in f
                if not line.startswith('#') and line.strip()
            )

            hmm_res = {}
            for scaffold, name, evalue in it:
                if scaffold not in hmm_res:
                    hmm_res[scaffold] = {}
                hmm_res[scaffold][name] = 1 if evalue < max_evalue else 0

        with open(self.gene_mark.output()) as f:
            with open(self.output(), 'w') as of:
                writer = csv.writer(of, delimiter='\t')
                it = (
                    line.strip().split('\t')
                    for line in f
                    if line.startswith('>')
                )
                # for i in it:
                #     print(i)

                for scaffold, group in groupby(it, key=lambda t: t[1]):
                    if scaffold not in hmm_res:
                        continue
                    scaffold_res = hmm_res[scaffold]

                    is_phage_it = (
                        scaffold_res.get(name[1:], 0)
                        for name, _ in group
                    )

                    is_phage_it = (
                        'P' if is_phage else 'N'
                        for is_phage in is_phage_it
                    )
                    writer.writerow((scaffold, ''.join(is_phage_it)))




from builtins import super
import csv
from itertools import groupby

from phigaro.data import convert_npn
from phigaro.finder.v2 import V2Finder
from .base import AbstractTask
from .hmmer import HmmerTask
from .gene_mark import GeneMarkTask


class RunPhigaroTask(AbstractTask):
    task_name = 'run_phigaro'

    def __init__(self, hmmer_task, gene_mark_task):
        """

        :type hmmer_task: HmmerTask
        :type gene_mark_task: GeneMarkTask
        """
        super().__init__()
        self.hmmer_task = hmmer_task
        self.gene_mark_task = gene_mark_task

    def _prepare(self):
        self.finder = V2Finder(
            window_len=self.config['phigaro']['window_len'],
            threshold_min=self.config['phigaro']['threshold_min'],
            threshold_max=self.config['phigaro']['threshold_max'],
        )

    def output(self):
        return self.file('{}.tsv'.format(self.sample))

    def read_gene_coords(self):
        def extract_coords(gene_str):
            tokens = gene_str.split('|')
            return int(tokens[-2]), int(tokens[-1])

        with open(self.gene_mark_task.output()) as f:
            genes_scaffolds = (
                line.strip().split('\t')
                for line in f
                if line.startswith('>')
            )

            genes_scaffolds = (
                (scaffold, extract_coords(gene_str))
                for gene_str, scaffold in genes_scaffolds
            )

            return {
                scaffold: [
                    coords
                    for _, coords in group
                ]
                for scaffold, group in groupby(genes_scaffolds, key=lambda x: x[0])
            }

    def run(self):
        scaffold_genes_coords = self.read_gene_coords()

        with open(self.output(), 'w') as of:
            writer = csv.writer(of, delimiter='\t')

            with open(self.hmmer_task.output()) as f:
                reader = csv.reader(f, delimiter='\t')
                for scaffold, npn_str in reader:
                    genes_coords = scaffold_genes_coords[scaffold]

                    npn = convert_npn(npn_str, 'P')
                    phages = self.finder.find_phages(npn)
                    for phage in phages:
                        begin = genes_coords[phage.begin][0]
                        end = genes_coords[phage.end][1]
                        writer.writerow((scaffold, begin, end))

